from typing import Optional, Union

import pandas as pd

from ...base import WeightingStrategy
from ..utils import calculate_rolling_window_size


class WeightByMax(WeightingStrategy):
    def __init__(self, window_size: Optional[Union[int, float]]):
        self.window_size = window_size

    def calculate(self, series: pd.Series) -> pd.Series:
        return calculate_rolling_expanding_abs_max(series, self.window_size)

    def __str__(self) -> str:
        return f"WeightByMax-{self.window_size}"


class WeightByMaxWithLookahead(WeightingStrategy):
    def calculate(self, series: pd.Series) -> pd.Series:
        maximum = series.max()
        sample_weights = series.abs() / maximum
        return sample_weights.fillna(0.0)

    def __str__(self) -> str:
        return "WeightByMaxWithLookahead"


class WeightBySumWithLookahead(WeightingStrategy):
    def calculate(self, series: pd.Series) -> pd.Series:
        sample_weights = series.abs() / series.sum()
        return sample_weights.fillna(0.0)

    def __str__(self) -> str:
        return "WeightBySumWithLookahead"


def calculate_rolling_expanding_abs_max(
    series: pd.Series, window_size: Optional[Union[int, float]]
) -> pd.Series:
    abs_returns = series.abs()
    rolling_or_expanding = (
        abs_returns.expanding()
        if window_size is None
        else abs_returns.rolling(
            calculate_rolling_window_size(window_size, abs_returns), min_periods=0
        )
    )
    maximum = rolling_or_expanding.max()
    sample_weights = series.abs() / maximum
    return sample_weights.fillna(0.0)



from typing import Callable, List, Tuple, Union

from ..utils.list import empty_if_none, flatten
from .classes import (
    Composite,
    Optimizer,
    Pipeline,
    Pipelines,
    Sampler,
    TrainedPipeline,
    TrainedPipelines,
    Transformation,
)


def traverse_apply(pipeline: Pipeline, apply_func: Callable) -> Pipeline:
    def _traverse_apply(pipeline):
        if isinstance(pipeline, List) or isinstance(pipeline, Tuple):
            return [_traverse_apply(t) for t in pipeline]
        else:
            return apply_func(pipeline, _traverse_apply)

    return _traverse_apply(pipeline)


def traverse(
    pipeline: Pipeline,
):
    """
    Iterates over a pipeline and yields each transformation.
    CAUTION: It does not "unroll" Optimizer's candidates.
    """
    if isinstance(pipeline, List) or isinstance(pipeline, Tuple):
        for i in pipeline:
            yield from traverse(i)
    elif isinstance(pipeline, Composite):
        yield pipeline
        items = pipeline.get_children_primary() + empty_if_none(
            pipeline.get_children_secondary()
        )
        for i in items:
            yield from traverse(i)
    elif isinstance(pipeline, Sampler):
        yield pipeline
        yield from traverse(pipeline.get_children_primary())
    elif isinstance(pipeline, Optimizer):
        yield pipeline
        if pipeline.get_optimized_pipeline() is not None:
            yield pipeline.get_optimized_pipeline()
        else:
            yield pipeline.get_candidates()[0]
    else:
        yield pipeline


def get_flat_list_of_transformations(
    transformations: Pipelines,
) -> List[Transformation]:
    return flatten(
        [t for t in traverse(transformations) if isinstance(t, Transformation)]
    )


def get_concatenated_names(transformations: Union[Pipelines, Pipeline]) -> str:
    return "-".join(
        [
            transformation.name if hasattr(transformation, "name") else ""
            for transformation in get_flat_list_of_transformations(transformations)
        ]
    )


def get_maximum_memory_size(pipeline: Pipeline) -> int:
    memory_sizes = [
        t.properties.memory_size
        for t in get_flat_list_of_transformations(pipeline)
        if t.properties.memory_size is not None
    ]
    if len(memory_sizes) == 0:
        return 0
    else:
        return max(memory_sizes)


def get_last_trained_pipeline(trained_pipelines: TrainedPipelines) -> TrainedPipeline:
    return [series.iloc[-1] for series in trained_pipelines]



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


import collections
from collections.abc import Iterable
from typing import Dict, List, Optional, Tuple, TypeVar, Union

from iteration_utilities import unique_everseen

T = TypeVar("T")


def wrap_in_list(input: Union[T, List[T]]) -> List[T]:
    return input if isinstance(input, List) else [input]


def transform_range_to_list(input: Union[range, List[T]]) -> List[T]:
    return list(input) if isinstance(input, range) else input


def wrap_in_double_list_if_needed(
    input: Union[T, List[T]]
) -> Union[List[List[T]], List[T]]:
    """
    If input is a single item, wrap it in a list.
    If input is a single list, wrap it in another list.
    If input is a list of lists, return it as is.
    """
    if not isinstance(input, List):
        return [input]
    if isinstance(input[0], List):
        return input
    else:
        return [input]


def flatten(input: Union[List[List], List]) -> Iterable:
    for x in input:
        if isinstance(x, list):
            yield from flatten(x)
        else:
            yield x


def keep_only_duplicates(input: List) -> List:
    return [item for item, count in collections.Counter(input).items() if count > 1]


def has_intersection(lhs: List, rhs: List) -> bool:
    return len(set(lhs).intersection(rhs)) > 0


def unique(input: List) -> List:
    return unique_everseen(input)


def swap_tuples(input: List[Tuple]) -> List[Tuple]:
    return [(b, a) for a, b in input]


def filter_none(input: List) -> List:
    return [x for x in input if x is not None]


def empty_if_none(input: Union[List, None]) -> List:
    return [] if input is None else input


def to_hierachical_dict(flat_dict: dict, seperator: str = "¦") -> dict:
    hierarchy = {}
    for key, value in flat_dict.items():
        name_and_param = key.split(seperator)
        unraveled_obj_key = name_and_param[0]
        unraveled_param_key = name_and_param[1]
        if unraveled_obj_key not in hierarchy:
            hierarchy[unraveled_obj_key] = {}
        hierarchy[unraveled_obj_key][unraveled_param_key] = value
    return hierarchy


def to_hierachical_dict_arbitrary_depth(flat_dict: dict, separator: str = "¦") -> dict:
    hierarchy = {}
    for key, value in flat_dict.items():
        current_dict = hierarchy
        parts = key.split(separator)
        for part in parts[:-1]:
            if part not in current_dict:
                current_dict[part] = {}
            current_dict = current_dict[part]
        current_dict[parts[-1]] = value
    return hierarchy


def ensure_dict(dictionary: Optional[Dict]) -> dict:
    return {} if dictionary is None else dictionary


def unpack_list_of_tuples(input: List[Tuple]):
    if len(input) == 1:
        return [[item] for item in input[0]]
    return zip(*input)



from __future__ import annotations

import enum
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Callable, List, Optional, Tuple, TypeVar, Union

import pandas as pd
from typing_extensions import Self

from ..splitters import SingleWindowSplitter
from ..utils.dataframe import ResolutionStrategy, concat_on_columns_with_duplicates
from ..utils.enums import ParsableEnum
from ..utils.introspection import get_initialization_parameters

T = TypeVar("T", Optional[pd.Series], pd.Series)
X = pd.DataFrame


class Block(ABC):
    id: str
    name: str

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        instance.id = str(uuid.uuid4())
        return instance


class Clonable(ABC):
    @abstractmethod
    def clone(self, clone_children: Callable) -> Self:
        raise NotImplementedError


class Composite(Block, Clonable, ABC):
    """
    A Composite contains other transformations.
    """

    @dataclass
    class Properties:
        primary_requires_predictions: bool = (
            False  # Primary pipeline need output from a model
        )
        primary_only_single_pipeline: bool = (
            False  # There should be a single primary pipeline.
        )
        secondary_requires_predictions: bool = (
            False  # Secondary pipeline need output from a model
        )
        secondary_only_single_pipeline: bool = (
            False  # There should be a single secondary pipeline.
        )
        artifacts_length_should_match: bool = (
            True  # returned Artifacts should be the same length as the input
        )

    @dataclass
    class Metadata:
        project_name: str
        fold_index: int
        target: str
        inference: bool

    properties: Properties
    metadata: Optional[Metadata]

    @abstractmethod
    def get_children_primary(self) -> Pipelines:
        raise NotImplementedError

    def get_children_secondary(
        self,
    ) -> Optional[Pipelines]:
        return None

    def postprocess_result_primary(
        self,
        results: List[pd.DataFrame],
        y: Optional[pd.Series],
        original_artifact: Artifact,
        fit: bool,
    ) -> pd.DataFrame:
        raise NotImplementedError

    def postprocess_result_secondary(
        self,
        primary_results: pd.DataFrame,
        secondary_results: List[pd.DataFrame],
        y: Optional[pd.Series],
        in_sample: bool,
    ) -> pd.DataFrame:
        raise NotImplementedError

    def before_fit(self, X: pd.DataFrame) -> None:
        pass

    def preprocess_primary(
        self, X: pd.DataFrame, index: int, y: T, artifact: Artifact, fit: bool
    ) -> Tuple[pd.DataFrame, T, Artifact]:
        return X, y, artifact

    def preprocess_secondary(
        self,
        X: pd.DataFrame,
        y: T,
        artifact: Artifact,
        results_primary: pd.DataFrame,
        index: int,
        fit: bool,
    ) -> Tuple[pd.DataFrame, T, Artifact]:
        return X, y, artifact

    def postprocess_artifacts_primary(
        self,
        primary_artifacts: List[Artifact],
        results: List[pd.DataFrame],
        original_artifact: Artifact,
        fit: bool,
    ) -> pd.DataFrame:
        return concat_on_columns_with_duplicates(
            primary_artifacts, strategy=ResolutionStrategy.last
        )

    def postprocess_artifacts_secondary(
        self,
        primary_artifacts: pd.DataFrame,
        secondary_artifacts: List[Artifact],
        original_artifact: Artifact,
    ) -> pd.DataFrame:
        return concat_on_columns_with_duplicates(
            [primary_artifacts] + secondary_artifacts, strategy=ResolutionStrategy.last
        )


class Sampler(Block, Clonable, ABC):
    @abstractmethod
    def get_children_primary(self) -> Pipelines:
        raise NotImplementedError

    def before_fit(self, X: pd.DataFrame) -> None:
        pass

    def preprocess_primary(
        self, X: pd.DataFrame, index: int, y: T, artifact: Artifact, fit: bool
    ) -> Tuple[pd.DataFrame, T, Artifact]:
        return X, y, artifact


class Optimizer(Block, Clonable, ABC):
    splitter: SingleWindowSplitter

    @abstractmethod
    def get_candidates(self) -> List[Pipeline]:
        """
        Called iteratively, until an array with a length of zero is returned.
        Then the loop finishes the candidate evaluation process.
        """
        raise NotImplementedError

    @abstractmethod
    def get_optimized_pipeline(self) -> Optional[Pipeline]:
        raise NotImplementedError

    @abstractmethod
    def process_candidate_results(
        self,
        results: List[pd.DataFrame],
        y: pd.Series,
        artifacts: List[pd.DataFrame],
    ) -> Optional[Artifact]:
        raise NotImplementedError


class Transformation(Block, ABC):
    """
    A transformation is a single step in a pipeline.
    """

    @dataclass
    class Properties:
        class ModelType(enum.Enum):
            regressor = "regressor"
            classifier = "classifier"

        class Mode(enum.Enum):
            minibatch = "minibatch"
            online = "online"

        requires_X: bool
        mode: Mode = Mode.minibatch
        updateable: bool = True
        memory_size: Optional[
            Union[int, float]
        ] = None  # During the in_sample period, memory will contain all data. During inference, if not `None`, it will inject past window with size of `memory` to update() & transformation().
        disable_memory: bool = False  # If `True`, memory will not be used at all, even if `memory_size` is not `None`.
        model_type: Optional[ModelType] = None
        _internal_supports_minibatch_backtesting: bool = False  # internal, during backtesting, calls predict_in_sample() instead of predict()

    @dataclass
    class State:
        memory_X: pd.DataFrame
        memory_y: pd.Series
        memory_sample_weights: Optional[pd.Series]

    properties: Properties
    name: str
    _state: Optional[State]

    @abstractmethod
    def fit(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        sample_weights: Optional[pd.Series] = None,
    ) -> Optional[Artifact]:
        """
        Called once, with on initial training window.
        """
        raise NotImplementedError

    @abstractmethod
    def update(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        sample_weights: Optional[pd.Series] = None,
    ) -> Optional[Artifact]:
        """
        Subsequent calls to update the model.
        """
        raise NotImplementedError

    @abstractmethod
    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        raise NotImplementedError


class InvertibleTransformation(Transformation, ABC):
    @abstractmethod
    def inverse_transform(self, X: pd.Series, in_sample: bool) -> pd.Series:
        raise NotImplementedError


class Tunable(ABC):
    params_to_try: Optional[dict]

    def get_params(self) -> dict:
        """
        The default implementation assumes that:
        1. All init parameters are stored on the object as property (with the same name/key).
        2. There are no modifications/conversions of the init parameters that'd prevent them from being used again (reconstructing the object from them).
        """
        return get_initialization_parameters(self)

    def get_params_to_try(self) -> Optional[dict]:
        return self.params_to_try

    def clone_with_params(
        self, parameters: dict, clone_children: Optional[Callable] = None
    ) -> Tunable:
        """
        The default implementation only works for Transformations, when parameters and the init parameters match 100%.
        """
        return self.__class__(**parameters)


class FeatureSelector(Transformation):
    selected_features: List[str]


Pipeline = Union[Block, List[Block]]
"""A list of `fold` objects that are executed sequentially. Or a single object."""
Pipelines = List[Pipeline]
"""Multiple, independent `Pipeline`s."""

TransformationPipeline = Union[
    Union[Transformation, Composite], List[Union[Transformation, Composite]]
]


TrainedPipeline = Pipeline
TrainedPipelines = List[pd.Series]
"""A list of trained `Pipeline`s, to be used for backtesting."""
OutOfSamplePredictions = pd.DataFrame
"""The backtest's resulting out-of-sample output."""
InSamplePredictions = pd.DataFrame
"""The backtest's resulting in-sample output."""


@dataclass
class PipelineCard:
    preprocessing: Optional[TransformationPipeline]
    pipeline: Pipeline
    event_labeler: Optional[Labeler] = None
    event_filter: Optional[EventFilter] = None
    name: Optional[str] = None


@dataclass
class TrainedPipelineCard:
    name: str
    preprocessing: TrainedPipelines
    pipeline: TrainedPipelines
    event_labeler: Optional[Labeler]
    event_filter: Optional[EventFilter]


def fit_noop(
    self,
    X: pd.DataFrame,
    y: pd.Series,
    sample_weights: Optional[pd.Series] = None,
) -> None:
    pass


class SingleFunctionTransformation(Transformation):
    properties = Transformation.Properties(requires_X=True)

    def get_function(self) -> Callable:
        raise NotImplementedError

    fit = fit_noop
    update = fit_noop

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        return pd.concat([X, self.get_function()(X)], axis="columns")


class EventDataFrame(pd.DataFrame):
    @property
    def start(self) -> pd.Series:
        return self["event_start"]

    @property
    def end(self) -> pd.Series:
        return self["event_end"]

    @property
    def label(self) -> pd.Series:
        return self["event_label"]

    @property
    def raw(self) -> pd.Series:
        return self["event_raw"]

    @property
    def sample_weights(self) -> pd.Series:
        return self["event_sample_weights"]

    @property
    def test_sample_weights(self) -> pd.Series:
        return self["event_test_sample_weights"]

    @classmethod
    def from_data(
        cls,
        start: pd.DatetimeIndex,
        end: pd.DatetimeIndex,
        label: pd.Series,
        raw: pd.Series,
        sample_weights: Optional[pd.Series] = None,
        test_sample_weights: Optional[pd.Series] = None,
    ) -> EventDataFrame:
        return cls(
            data={
                "event_start": start,
                "event_end": end,
                "event_label": label,
                "event_raw": raw,
                "event_sample_weights": pd.Series(1.0, index=start)
                if sample_weights is None
                else sample_weights,
                "event_test_sample_weights": pd.Series(1.0, index=start)
                if test_sample_weights is None
                else test_sample_weights,
            }
        )


class Artifact(pd.DataFrame):
    @staticmethod
    def empty(index: pd.Index) -> Artifact:
        return pd.DataFrame(index=index)  # type: ignore

    @staticmethod
    def dummy_events(index: pd.Index) -> Artifact:
        return pd.DataFrame(
            index=index,
            data={
                "event_start": index,
                "event_end": index,
                "event_label": [1.0] * len(index),
                "event_raw": [9.0] * len(index),
                "event_sample_weights": [1.0] * len(index),
                "event_test_sample_weights": [1.0] * len(index),
            },
        )  # type: ignore

    @staticmethod
    def get_sample_weights(artifact: Artifact) -> Optional[pd.Series]:
        if "event_sample_weights" not in artifact.columns:
            return None
        return artifact["event_sample_weights"]

    @staticmethod
    def get_test_sample_weights(artifact: Artifact) -> Optional[pd.Series]:
        if "event_test_sample_weights" not in artifact.columns:
            if "sample_weights" in artifact.columns:
                return artifact["sample_weights"]
            else:
                return None
        return artifact["event_test_sample_weights"]

    @staticmethod
    def get_event_label(artifact: Artifact) -> Optional[pd.Series]:
        if "event_label" not in artifact.columns:
            return None
        return artifact["event_label"]

    @staticmethod
    def get_events(artifact: Artifact) -> Optional[EventDataFrame]:
        if "event_start" not in artifact.columns:
            return None
        return EventDataFrame.from_data(
            start=artifact.event_start,
            end=artifact.event_end,
            label=artifact.event_label,
            raw=artifact.event_raw,
            sample_weights=artifact.event_sample_weights,
            test_sample_weights=artifact.event_test_sample_weights,
        )

    @staticmethod
    def from_events(
        index: pd.Index,
        events: Optional[EventDataFrame],
    ) -> Artifact:
        if events is not None and not events.columns[0].startswith("event_"):
            events = events.add_prefix("event_")
        if events is None or events.empty:
            return Artifact.empty(index)  # type: ignore
        else:
            return events.reindex(index)  # type: ignore


class PredefinedFunction(ParsableEnum):
    mean = "mean"
    sum = "sum"
    median = "median"
    std = "std"
    var = "var"
    kurt = "kurt"
    min = "min"
    max = "max"
    corr = "corr"
    cov = "cov"
    skew = "skew"
    sem = "sem"


class EventFilter(ABC):
    @abstractmethod
    def get_event_start_times(self, y: pd.Series) -> pd.DatetimeIndex:
        raise NotImplementedError


class Labeler(ABC):
    @abstractmethod
    def label_events(
        self, event_start_times: pd.DatetimeIndex, y: pd.Series
    ) -> EventDataFrame:
        raise NotImplementedError

    def get_all_possible_labels(self) -> List[int]:
        raise NotImplementedError


class LabelingStrategy(ABC):
    @abstractmethod
    def label(self, series: pd.Series) -> pd.Series:
        raise NotImplementedError

    @abstractmethod
    def get_all_labels(self) -> List[int]:
        raise NotImplementedError


class WeightingStrategy(ABC):
    @abstractmethod
    def calculate(self, series: pd.Series) -> pd.Series:
        raise NotImplementedError



import importlib
from typing import TYPE_CHECKING, Optional, Tuple

import pandas as pd

from ..utils.checks import (
    all_have_probabilities,
    get_prediction_column,
    get_probabilities_columns,
)
from .classes import Artifact, OutOfSamplePredictions

if TYPE_CHECKING and importlib.util.find_spec("krisi") is not None:
    from krisi import ScoreCard


def score_results(
    result: pd.DataFrame,
    y: pd.Series,
    artifacts: Artifact,
    krisi_args: Optional[dict] = None,
) -> Tuple["ScoreCard", OutOfSamplePredictions]:
    y, pred_point, probabilities, test_sample_weights = align_result_with_events(
        y=y,
        result=result,
        artifacts=artifacts,
    )
    if importlib.util.find_spec("krisi") is not None:
        from krisi import score

        return (
            score(
                y=y,
                predictions=pred_point,
                probabilities=probabilities,
                sample_weight=test_sample_weights
                if test_sample_weights is not None
                else None,
                **(krisi_args if krisi_args is not None else {}),
            ),
            pd.concat([pred_point, probabilities], axis="columns"),
        )
    else:
        raise ImportError(
            "krisi not installed. Please install krisi to use this function."
        )


def align_result_with_events(
    y: pd.Series,
    result: pd.DataFrame,
    artifacts: Artifact,
) -> Tuple[pd.Series, pd.Series, Optional[pd.DataFrame], Optional[pd.Series]]:
    events = Artifact.get_events(artifacts)
    test_sample_weights = Artifact.get_test_sample_weights(artifacts)

    probabilities = (
        get_probabilities_columns(result) if all_have_probabilities([result]) else None
    )
    pred_point = get_prediction_column(result)

    if events is not None:
        events = events.reindex(result.index).dropna()
        y = events.event_label
        test_sample_weights = events.event_test_sample_weights
        pred_point = pred_point.dropna()
        probabilities = probabilities.dropna() if probabilities is not None else None
        y = y[pred_point.index]
    else:
        test_sample_weights = (
            test_sample_weights[pred_point.index]
            if test_sample_weights is not None
            else None
        )
        y = y[pred_point.index]

    return y, pred_point, probabilities, test_sample_weights



from __future__ import annotations

from logging import getLogger
from typing import Callable, List, Optional, Union

import pandas as pd

from ...base import (
    EventDataFrame,
    Labeler,
    LabelingStrategy,
    PredefinedFunction,
    WeightingStrategy,
)
from ...utils.forward import create_forward_rolling
from ..weights import NoWeighting, WeightBySumWithLookahead

logger = getLogger("fold:labeling")


class FixedForwardHorizon(Labeler):
    time_horizon: int

    def __init__(
        self,
        time_horizon: int,
        labeling_strategy: LabelingStrategy,
        weighting_strategy: Optional[WeightingStrategy],
        weighting_strategy_test: WeightingStrategy = WeightBySumWithLookahead(),
        transformation_function: Optional[Callable] = None,
        aggregate_function: Union[
            str, PredefinedFunction, Callable
        ] = PredefinedFunction.sum,
        truncate_if_timeframe_is_smaller: bool = True,
        flip_sign: bool = False,
        shift_by: Optional[int] = None,
    ):
        self.time_horizon = time_horizon
        self.labeling_strategy = labeling_strategy
        self.weighting_strategy = (
            weighting_strategy if weighting_strategy else NoWeighting()
        )
        self.weighting_strategy_test = weighting_strategy_test
        self.transformation_function = transformation_function
        self.aggregate_function = (
            aggregate_function
            if isinstance(aggregate_function, Callable)
            else getattr(
                pd.core.window.rolling.Rolling,
                PredefinedFunction.from_str(aggregate_function).value,
            )
        )
        self.flip_sign = flip_sign
        self.shift_by = shift_by
        self.truncate_if_timeframe_is_smaller = truncate_if_timeframe_is_smaller

    def label_events(
        self, event_start_times: pd.DatetimeIndex, y: pd.Series
    ) -> EventDataFrame:
        forward_rolling_aggregated = create_forward_rolling(
            transformation_func=self.transformation_function,
            agg_func=self.aggregate_function,
            series=y,
            period=self.time_horizon,
            shift_by=self.shift_by,
            truncate_end=self.truncate_if_timeframe_is_smaller,
        )
        if self.truncate_if_timeframe_is_smaller:
            cutoff_point = y.index[-self.time_horizon]
            event_start_times = event_start_times[event_start_times < cutoff_point]

        forward_rolling_aggregated = forward_rolling_aggregated[
            event_start_times
        ]  # filter based on events

        labels = self.labeling_strategy.label(forward_rolling_aggregated)
        sample_weights = self.weighting_strategy.calculate(forward_rolling_aggregated)
        test_sample_weights = self.weighting_strategy_test.calculate(
            forward_rolling_aggregated
        )

        if self.flip_sign:
            if len(labels.dropna().unique()) == 3:
                labels = labels * -1
            elif len(labels.dropna().unique()) == 2:
                labels = 1 - labels
            else:
                logger.warn(
                    f"{len(labels.dropna().unique())} classes detected, can't flip sign"
                )

        offset = pd.Timedelta(value=self.time_horizon, unit=y.index.freqstr)
        return EventDataFrame.from_data(
            start=event_start_times,
            end=(event_start_times + offset).astype("datetime64[s]"),
            label=labels,
            raw=forward_rolling_aggregated,
            sample_weights=sample_weights,
            test_sample_weights=test_sample_weights,
        )

    def get_labels(self) -> List[int]:
        return self.labeling_strategy.get_all_labels()



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from __future__ import annotations

from typing import Optional, Tuple

import pandas as pd

from ...base import Artifact, Transformation, X
from ...utils.checks import is_X_available
from ...utils.dataframe import concat_on_columns
from ..memory import postprocess_X_y_into_memory_, preprocess_X_y_with_memory
from ..types import Stage


def _process_minibatch_transformation(
    transformation: Transformation,
    X: pd.DataFrame,
    y: Optional[pd.Series],
    artifacts: Artifact,
    stage: Stage,
) -> Tuple[Transformation, X, Artifact]:
    if not is_X_available(X) and transformation.properties.requires_X:
        raise ValueError(
            "X is None, but transformation"
            f" {transformation.__class__.__name__} requires it."
        )

    in_sample = stage == Stage.inital_fit
    (
        X_with_memory,
        y_with_memory,
        sample_weights_with_memory,
    ) = preprocess_X_y_with_memory(
        transformation,
        X,
        y,
        Artifact.get_sample_weights(artifacts),
        in_sample=in_sample,
    )
    # The order is:
    # 1. fit (if we're in the initial_fit stage)
    artifact = None
    if stage == Stage.inital_fit:
        artifact = transformation.fit(
            X_with_memory, y_with_memory, sample_weights_with_memory
        )
        postprocess_X_y_into_memory_(
            transformation,
            X_with_memory,
            y_with_memory,
            sample_weights_with_memory,
            in_sample=stage == Stage.inital_fit,
        )
        artifacts = concat_on_columns([artifact, artifacts], copy=False)
    # 2. transform (inference)
    (
        X_with_memory,
        y_with_memory,
        sample_weights_with_memory,
    ) = preprocess_X_y_with_memory(
        transformation,
        X,
        y,
        Artifact.get_sample_weights(artifacts),
        in_sample=False,
    )
    return_value = transformation.transform(X_with_memory, in_sample=in_sample)
    # 3. update (if we're in the update stage)
    if stage == Stage.update and transformation.properties.updateable:
        artifact = transformation.update(
            X_with_memory, y_with_memory, sample_weights_with_memory
        )
        artifacts = concat_on_columns([artifact, artifacts], copy=False)
        postprocess_X_y_into_memory_(
            transformation,
            X,
            y,
            Artifact.get_sample_weights(artifacts),
            in_sample=False,
        )
    return transformation, return_value.loc[X.index], artifacts


def _process_internal_online_model_minibatch_inference_and_update(
    transformation: Transformation,
    X: pd.DataFrame,
    y: Optional[pd.Series],
    artifacts: Artifact,
) -> Tuple[Transformation, X, Artifact]:
    (
        X_with_memory,
        y_with_memory,
        sample_weights_with_memory,
    ) = preprocess_X_y_with_memory(
        transformation,
        X,
        y,
        Artifact.get_sample_weights(artifacts),
        in_sample=True,
    )
    postprocess_X_y_into_memory_(
        transformation,
        X_with_memory,
        y_with_memory,
        sample_weights_with_memory,
        in_sample=True,
    )
    return_value = transformation.transform(X_with_memory, in_sample=True)

    artifact = transformation.update(
        X_with_memory, y_with_memory, sample_weights_with_memory
    )
    postprocess_X_y_into_memory_(
        transformation,
        X,
        y,
        Artifact.get_sample_weights(artifacts),
        in_sample=False,
    )
    return (
        transformation,
        return_value.loc[X.index],
        concat_on_columns([artifact, artifacts], copy=False),
    )



# from fold.base import Composite
# from fold.composites.residual import ModelResiduals
# from fold.composites.target import TransformTarget
# from fold.transformations.dev import Identity
# from fold.transformations.difference import Difference

# from .ar import AR


# class ARMA(ModelResiduals):
#     """
#     ARMA(p, q) model.

#     Parameters
#     ----------
#     p : int
#         The number of lags to include in the model.
#     q : int
#         The number of lags to include in the residual model.
#     ma_model_online : bool, optional
#         Whether to use the online version of the MA model, by default False.
#     """

#     properties = Composite.Properties(
#         primary_requires_predictions=True,
#         primary_only_single_pipeline=True,
#         secondary_requires_predictions=False,
#         secondary_only_single_pipeline=True,
#     )

#     def __init__(self, p: int, q: int, ma_model_online: bool = False):
#         assert p >= 0, "p must be above 0"
#         assert q >= 0, "q must be above 0"

#         if q > 0:
#             ma_model = AR(q)
#             ma_model.properties._internal_supports_minibatch_backtesting = (
#                 not ma_model_online
#             )
#         else:
#             ma_model = Identity()
#         super().__init__(
#             primary=AR(p),
#             meta=ma_model,
#         )


# class ARIMA(TransformTarget):
#     """
#     ARIMA(p, d, q) model.

#     Parameters
#     ----------
#     p : int
#         The number of lags to include in the model.
#     d : int
#         The number of times to difference the target.
#     q : int
#         The number of lags to include in the residual model.
#     ma_model_online : bool, optional
#         Whether to use the online version of the MA model, by default False.
#     """

#     def __init__(self, p: int, d: int, q: int, ma_model_online: bool = True):
#         assert p >= 0, "p must be above 0"
#         assert q >= 0, "q must be above 0"
#         assert d >= 0, "d must be above 0"

#         arma = ARMA(p, q, ma_model_online)
#         differencers = Identity() if d == 0 else [Difference(1) for _ in range(0, d)]

#         super().__init__(wrapped_pipeline=arma, y_pipeline=differencers)



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from __future__ import annotations

from typing import Optional, Tuple

import pandas as pd

from ...base import Artifact, Transformation, X
from ...utils.dataframe import concat_on_columns, concat_on_index
from ..memory import postprocess_X_y_into_memory_, preprocess_X_y_with_memory


def _process_with_inner_loop(
    transformation: Transformation,
    X: pd.DataFrame,
    y: Optional[pd.Series],
    artifacts: Artifact,
) -> Tuple[Transformation, X, Artifact]:
    if len(X) == 0:
        return (transformation, pd.DataFrame(), artifacts)

    # We need to run the inference & fit loop on each row, sequentially (one-by-one).
    # This is so the transformation can update its parameters after each sample.

    def transform_row(
        X_row: pd.DataFrame,
        y_row: Optional[pd.Series],
        artifact_row: Artifact,
    ):
        (
            X_row_with_memory,
            y_row_with_memory,
            sample_weights_row_with_memory,
        ) = preprocess_X_y_with_memory(
            transformation,
            X_row,
            y_row,
            Artifact.get_sample_weights(artifact_row),
            in_sample=False,
        )
        result = transformation.transform(X_row_with_memory, in_sample=False)
        if y_row is not None:
            artifact = transformation.update(
                X_row_with_memory, y_row_with_memory, sample_weights_row_with_memory
            )
            _ = concat_on_columns([artifact, artifacts], copy=False)
            postprocess_X_y_into_memory_(
                transformation,
                X_row_with_memory,
                y_row_with_memory,
                sample_weights_row_with_memory,
                in_sample=False,
            )
        return result.loc[X_row.index]

    return (
        transformation,
        concat_on_index(
            [
                transform_row(
                    X.loc[index:index],
                    y.loc[index:index] if y is not None else None,
                    artifacts[index:index],
                )
                for index in X.index
            ],
            copy=False,
        ),
        Artifact.empty(X.index),
    )



import inspect


def get_initialization_parameters(instance):
    init_args = inspect.signature(instance.__init__).parameters.keys()
    return {key: value for key, value in instance.__dict__.items() if key in init_args}



from typing import List

import pandas as pd

from ...base import LabelingStrategy


class BinarizeSign(LabelingStrategy):
    def label(self, series: pd.Series) -> pd.Series:
        labels = series.copy()
        labels.loc[series >= 0.0] = 1
        labels.loc[series < 0.0] = 0
        return labels

    def get_all_labels(self) -> List[int]:
        return [0, 1]


class NoLabel(LabelingStrategy):
    def label(self, series: pd.Series) -> pd.Series:
        return series

    def get_all_labels(self) -> List[int]:
        return [0, 1]



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from __future__ import annotations

from typing import List, Optional, Union

import pandas as pd

from ..base import Transformation, Tunable, fit_noop
from ..utils.checks import get_list_column_names
from ..utils.list import wrap_in_list


class SelectColumns(Transformation, Tunable):
    """
    Selects a single or multiple columns, drops the rest.

    Parameters
    ----------

    columns : Union[List[str], str]
        The column or columns to select (dropping the rest).


    """

    def __init__(
        self,
        columns: Union[List[str], str],
        name: Optional[str] = None,
        params_to_try: Optional[dict] = None,
    ) -> None:
        self.columns: List[str] = wrap_in_list(columns)
        self.params_to_try = params_to_try
        self.name = name or f"SelectColumns-{columns}"
        self.properties = Transformation.Properties(requires_X=True)

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        return X[get_list_column_names(self.columns, X)]

    fit = fit_noop
    update = fit


class DropColumns(Transformation, Tunable):
    """
    Drops a single or multiple columns.

    Parameters
    ----------

    columns : List[str], str
        The column or columns to drop.

    """

    def __init__(
        self,
        columns: Union[List[str], str],
        name: Optional[str] = None,
        params_to_try: Optional[dict] = None,
    ) -> None:
        self.columns = wrap_in_list(columns)
        self.params_to_try = params_to_try
        self.name = name or f"DropColumns-{columns}"
        self.properties = Transformation.Properties(requires_X=True)

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        return X.drop(columns=get_list_column_names(self.columns, X), errors="ignore")

    fit = fit_noop
    update = fit


class RenameColumns(Transformation):
    """
    Renames columns.

    Parameters
    ----------

    columns_mapper : dict
        A dictionary containing the old column names as keys and the new column names as values.

    Examples
    --------

    ```pycon
    >>> from fold.loop import train_backtest
    >>> from fold.splitters import SlidingWindowSplitter
    >>> from fold.transformations import RenameColumns
    >>> from fold.utils.tests import generate_sine_wave_data
    >>> X, y  = generate_sine_wave_data()
    >>> splitter = SlidingWindowSplitter(train_window=0.5, step=0.2)
    >>> pipeline = RenameColumns({"sine": "sine_renamed"})
    >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)
    >>> preds.head()
                         sine_renamed
    2021-12-31 15:40:00       -0.0000
    2021-12-31 15:41:00        0.0126
    2021-12-31 15:42:00        0.0251
    2021-12-31 15:43:00        0.0377
    2021-12-31 15:44:00        0.0502

    ```
    """

    def __init__(self, columns_mapper: dict) -> None:
        self.columns_mapper = columns_mapper
        self.name = "RenameColumns"
        self.properties = Transformation.Properties(requires_X=True)

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        return X.rename(columns=self.columns_mapper)

    fit = fit_noop
    update = fit


class AddColumnSuffix(Transformation):
    """
    Add suffix to column names.

    Parameters
    ----------

    suffix : str
        Suffix to add to all column names.

    Examples
    --------

    ```pycon
    >>> from fold.loop import train_backtest
    >>> from fold.splitters import SlidingWindowSplitter
    >>> from fold.transformations import RenameColumns
    >>> from fold.utils.tests import generate_sine_wave_data
    >>> X, y  = generate_sine_wave_data()
    >>> splitter = SlidingWindowSplitter(train_window=0.5, step=0.2)
    >>> pipeline = AddColumnSuffix("_2")
    >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)
    >>> preds.head()
                         sine_2
    2021-12-31 15:40:00 -0.0000
    2021-12-31 15:41:00  0.0126
    2021-12-31 15:42:00  0.0251
    2021-12-31 15:43:00  0.0377
    2021-12-31 15:44:00  0.0502

    ```
    """

    def __init__(self, suffix: str) -> None:
        self.suffix = suffix
        self.name = f"AddColumnSuffix-{self.suffix}"
        self.properties = Transformation.Properties(requires_X=True)

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        return X.add_suffix(self.suffix)

    fit = fit_noop
    update = fit


class OnlyPredictions(Transformation):
    """
    Drops all columns except the output model(s)' predictions.

    Examples
    --------
    ```pycon
    >>> from fold.loop import train_backtest
    >>> from fold.splitters import SlidingWindowSplitter
    >>> from fold.transformations import OnlyProbabilities
    >>> from fold.models.dummy import DummyClassifier
    >>> from fold.utils.tests import generate_sine_wave_data
    >>> X, y  = generate_sine_wave_data()
    >>> splitter = SlidingWindowSplitter(train_window=0.5, step=0.2)
    >>> pipeline = [DummyClassifier(1, [0, 1], [0.5, 0.5]), OnlyPredictions()]
    >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)
    >>> preds.head()
                         predictions_DummyClassifier
    2021-12-31 15:40:00                            1
    2021-12-31 15:41:00                            1
    2021-12-31 15:42:00                            1
    2021-12-31 15:43:00                            1
    2021-12-31 15:44:00                            1

    ```
    """

    def __init__(self) -> None:
        self.name = "OnlyPredictions"
        self.properties = Transformation.Properties(requires_X=True)

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        return X.drop(
            columns=[col for col in X.columns if not col.startswith("predictions_")]
        )

    fit = fit_noop
    update = fit


class OnlyProbabilities(Transformation):
    """
    Drops all columns except the output model(s)' probabilities.

    Examples
    --------
    ```pycon
    >>> from fold.loop import train_backtest
    >>> from fold.splitters import SlidingWindowSplitter
    >>> from fold.transformations import OnlyProbabilities
    >>> from fold.models.dummy import DummyClassifier
    >>> from fold.utils.tests import generate_sine_wave_data
    >>> X, y  = generate_sine_wave_data()
    >>> splitter = SlidingWindowSplitter(train_window=0.5, step=0.2)
    >>> pipeline = [DummyClassifier(1, [0, 1], [0.5, 0.5]), OnlyProbabilities()]
    >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)
    >>> preds.head()
                         probabilities_DummyClassifier_0  probabilities_DummyClassifier_1
    2021-12-31 15:40:00                              0.5                              0.5
    2021-12-31 15:41:00                              0.5                              0.5
    2021-12-31 15:42:00                              0.5                              0.5
    2021-12-31 15:43:00                              0.5                              0.5
    2021-12-31 15:44:00                              0.5                              0.5

    ```
    """

    def __init__(self) -> None:
        self.name = "OnlyProbabilities"
        self.properties = Transformation.Properties(requires_X=True)

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        return X.drop(
            columns=[col for col in X.columns if not col.startswith("probabilities_")]
        )

    fit = fit_noop
    update = fit



from typing import Optional

import pandas as pd

from ..base import Artifact, EventDataFrame, TrainedPipelineCard
from ..utils.trim import trim_initial_nans
from .backend import get_backend
from .checks import check_types
from .common import deepcopy_pipelines, recursively_transform
from .types import BackendType, Stage


def update(
    pipelinecard: TrainedPipelineCard,
    X: Optional[pd.DataFrame],
    y: pd.Series,
    events: Optional[EventDataFrame] = None,
) -> TrainedPipelineCard:
    """
    Update a set of Transformations with new data.
    Returns a new set of Transformations, does not mutate the original.
    """
    X, y = check_types(X, y)
    artifact = Artifact.from_events(X.index, events)
    X, y, artifact = trim_initial_nans(X, y, artifact)
    backend = get_backend(BackendType.no)

    if pipelinecard.preprocessing is not None:
        preprocessing_pipelines, X, preprocessing_artifact = recursively_transform(
            X,
            None,
            artifact,
            deepcopy_pipelines(pipelinecard.preprocessing),
            stage=Stage.update,
            backend=backend,
        )

    pipeline, result, artifact = recursively_transform(
        X,
        y,
        artifact,
        deepcopy_pipelines(pipelinecard.pipeline),
        stage=Stage.update,
        backend=backend,
    )
    return TrainedPipelineCard(
        name=pipelinecard.name,
        preprocessing=preprocessing_pipelines,
        pipeline=pipeline,
        event_filter=pipelinecard.event_filter,
        event_labeler=pipelinecard.event_labeler,
    )



def _wrap_xgboost(model):
    from xgboost import XGBClassifier, XGBRegressor, XGBRFClassifier, XGBRFRegressor

    from .gbd import WrapXGB

    if (
        isinstance(model, XGBRegressor)
        or isinstance(model, XGBRFRegressor)
        or isinstance(model, XGBClassifier)
        or isinstance(model, XGBRFClassifier)
    ):
        return WrapXGB.from_model(model)
    else:
        return None


def _wrap_lightgbm(model):
    from lightgbm import LGBMClassifier, LGBMRegressor

    from .gbd import WrapLGBM

    if isinstance(model, LGBMClassifier) or isinstance(model, LGBMRegressor):
        return WrapLGBM.from_model(model)
    else:
        return None


def _wrap_prophet(model):
    from prophet import Prophet

    from .prophet import WrapProphet

    if isinstance(model, Prophet):
        return WrapProphet.from_model(model)
    else:
        return None


def _wrap_sktime(model):
    from sktime.forecasting.base import BaseForecaster

    from .sktime import WrapSktime

    if isinstance(model, BaseForecaster):
        return WrapSktime.from_model(model)
    else:
        return None


def _wrap_statsforecast(model):
    from statsforecast.models import _TS

    from .statsforecast import WrapStatsForecast

    if isinstance(model, _TS):
        return WrapStatsForecast.from_model(model)
    else:
        return None



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from __future__ import annotations

from typing import Callable, List, Optional, Union

import pandas as pd

from fold.base.classes import Artifact

from ..base import Composite, Pipeline, Pipelines, get_concatenated_names
from ..transformations.columns import SelectColumns
from ..transformations.dev import Identity
from ..utils.dataframe import ResolutionStrategy, concat_on_columns_with_duplicates
from ..utils.list import wrap_in_double_list_if_needed, wrap_in_list


class Concat(Composite):
    """
    Concatenates the results of multiple pipelines.

    Parameters
    ----------

    pipelines : Pipelines
        A list of pipelines to be applied to the data, independently of each other.
    if_duplicate_keep : Union[ResolutionStrategy, str], optional
        How to handle duplicate columns, by default ResolutionStrategy.first
    custom_merge_logic: Optional[Callable[[List[pd.DataFrame]], pd.DataFrame]], default None
        A custom function that takes a list of dataframes and returns a single dataframe. If present, it's used instead of ResolutionStrategy.

    Examples
    --------
        >>> from fold.loop import train_backtest
        >>> from fold.splitters import SlidingWindowSplitter
        >>> from fold.composites import Concat
        >>> from fold.utils.tests import generate_sine_wave_data
        >>> X, y  = generate_sine_wave_data()
        >>> splitter = SlidingWindowSplitter(train_window=0.5, step=0.2)
        >>> pipeline = Concat([
        ...     lambda X: X.assign(sine_plus_1=X["sine"] + 1),
        ...     lambda X: X.assign(sine_plus_2=X["sine"] + 2),
        ... ])
        >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)
        >>> preds.head()
                             sine_plus_1  sine_plus_2    sine
        2021-12-31 15:40:00       1.0000       2.0000 -0.0000
        2021-12-31 15:41:00       1.0126       2.0126  0.0126
        2021-12-31 15:42:00       1.0251       2.0251  0.0251
        2021-12-31 15:43:00       1.0377       2.0377  0.0377
        2021-12-31 15:44:00       1.0502       2.0502  0.0502
    """

    ResolutionStrategy = ResolutionStrategy

    def __init__(
        self,
        pipelines: Pipelines,
        if_duplicate_keep: Union[ResolutionStrategy, str] = ResolutionStrategy.first,
        custom_merge_logic: Optional[
            Callable[[List[pd.DataFrame]], pd.DataFrame]
        ] = None,
        name: Optional[str] = None,
    ) -> None:
        self.pipelines = pipelines
        self.if_duplicate_keep = ResolutionStrategy.from_str(if_duplicate_keep)
        self.custom_merge_logic = custom_merge_logic
        self.name = name or "Concat-" + get_concatenated_names(pipelines)
        self.properties = Composite.Properties()
        self.metadata = None

    def postprocess_result_primary(
        self,
        results: List[pd.DataFrame],
        y: Optional[pd.Series],
        original_artifact: Artifact,
        fit: bool,
    ) -> pd.DataFrame:
        if self.custom_merge_logic is not None:
            return self.custom_merge_logic(results)
        else:
            return concat_on_columns_with_duplicates(results, self.if_duplicate_keep)

    def postprocess_artifacts_primary(
        self,
        primary_artifacts: List[Artifact],
        results: List[pd.DataFrame],
        original_artifact: Artifact,
        fit: bool,
    ) -> pd.DataFrame:
        return concat_on_columns_with_duplicates(
            primary_artifacts, self.if_duplicate_keep
        )

    def get_children_primary(self) -> Pipelines:
        return self.pipelines

    def clone(self, clone_children: Callable) -> Concat:
        clone = Concat(
            pipelines=clone_children(self.pipelines),
            if_duplicate_keep=self.if_duplicate_keep,
            custom_merge_logic=self.custom_merge_logic,
        )
        clone.properties = self.properties
        clone.name = self.name
        clone.metadata = self.metadata
        clone.id = self.id
        return clone


class Sequence(Composite):
    """
    An optional wrappers that is equivalent to using a single array for the transformations.
    It executes the transformations sequentially, in the order they are provided.

    Parameters
    ----------

    pipeline : Pipelines
        A list of transformations or models to be applied to the data.

    """

    def __init__(
        self,
        pipeline: Pipeline,
        name: Optional[str] = None,
    ) -> None:
        self.pipeline = wrap_in_double_list_if_needed(pipeline)
        self.name = name or "Sequence-" + get_concatenated_names(pipeline)
        self.properties = Composite.Properties(primary_only_single_pipeline=True)
        self.metadata = None

    def postprocess_result_primary(
        self,
        results: List[pd.DataFrame],
        y: Optional[pd.Series],
        original_artifact: Artifact,
        fit: bool,
    ) -> pd.DataFrame:
        return results[0]

    def get_children_primary(self) -> Pipelines:
        return self.pipeline

    def clone(self, clone_children: Callable) -> Sequence:
        clone = Sequence(pipeline=clone_children(self.pipeline))
        clone.properties = self.properties
        clone.name = self.name
        clone.metadata = self.metadata
        clone.id = self.id
        return clone


def TransformColumn(
    columns: Union[List[str], str],
    pipeline: Pipeline,
    name: Optional[str] = None,
) -> Composite:
    """
    Transforms a single or multiple columns using the given pipeline.
    """
    return Concat(
        [
            [SelectColumns(columns)] + wrap_in_list(pipeline),
            Identity(),
        ],
        if_duplicate_keep=ResolutionStrategy.first,
        name=name,
    )



import pandas as pd

from ...base import EventFilter


class NoFilter(EventFilter):
    def get_event_start_times(self, y: pd.Series) -> pd.DatetimeIndex:
        return y.index


class FilterZero(EventFilter):
    def get_event_start_times(self, y: pd.Series) -> pd.DatetimeIndex:
        return y[y != 0.0].index



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from ..base import *
from ..composites.concat import *
from .columns import *
from .date import *
from .difference import *
from .features import *
from .function import *
from .holidays import *
from .lags import *
from .math import *
from .scaling import *
from .sklearn import *



import pandas as pd

from fold.base.classes import Composite
from fold.loop.utils import _set_metadata

from ..base import (
    Artifact,
    OutOfSamplePredictions,
    TrainedPipelineCard,
    get_last_trained_pipeline,
    get_maximum_memory_size,
)
from .backend import get_backend
from .common import recursively_transform
from .types import BackendType, Stage


def infer(
    pipelinecard: TrainedPipelineCard,
    X: pd.DataFrame,
) -> OutOfSamplePredictions:
    """
    Run inference on a Pipeline and given data.
    Does not mutate or change the transformations in any way.
    A follow-up call to `update` is required to update the pipeline, when the ground truth is available.
    If X is None it will predict one step ahead.
    """
    assert isinstance(X, pd.DataFrame), "X must be a pandas DataFrame."
    last_pipeline = get_last_trained_pipeline(pipelinecard.pipeline)
    last_pipeline = _set_metadata(
        last_pipeline,
        Composite.Metadata(
            project_name=pipelinecard.name,
            fold_index=0,
            target="target",
            inference=True,
        ),
    )

    maximum_memory_size = max(
        get_maximum_memory_size(pipelinecard.preprocessing),
        get_maximum_memory_size(last_pipeline),
    )
    assert len(X) > maximum_memory_size, (
        f"X must be larger than {maximum_memory_size} rows, "
        f"but only {len(X)} rows were given."
    )

    if pipelinecard.preprocessing is not None:
        preprocessing = get_last_trained_pipeline(pipelinecard.preprocessing)
        preprocessing = _set_metadata(
            preprocessing,
            Composite.Metadata(
                project_name=f"{pipelinecard.name}-Preprocessing"
                if pipelinecard.name is not None
                else "Preprocessing",
                fold_index=0,
                target="target",
                inference=True,
            ),
        )
        _, X, _ = recursively_transform(
            X,
            None,
            Artifact.dummy_events(X.index),
            preprocessing,
            stage=Stage.infer,
            backend=get_backend(BackendType.no),
        )

    _, results, _ = recursively_transform(
        X,
        None,
        Artifact.dummy_events(X.index),
        last_pipeline,
        stage=Stage.infer,
        backend=get_backend(BackendType.no),
    )
    return results



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from __future__ import annotations

from typing import Callable, List, Optional, Tuple, Union

import pandas as pd

from ..base import (
    PredefinedFunction,
    Transformation,
    Tunable,
    feature_name_separator,
    fit_noop,
)
from ..utils.checks import get_list_column_names
from ..utils.dataframe import apply_function_batched, fill_na_inf
from ..utils.list import wrap_in_list

ColumnOrColumns = Union[str, List[str]]
FunctionOrPredefined = Union[Callable, PredefinedFunction, str]
ColumnWindowFunction = Tuple[ColumnOrColumns, Optional[int], FunctionOrPredefined]


class AddWindowFeatures(Transformation, Tunable):
    """
    Creates rolling window features on the specified columns.
    Equivalent to adding a new column by running: `df[column].rolling(window).function()`.


    Parameters
    ----------
    column_window_func : ColumnWindowFunction, List[ColumnWindowFunction]
        A list of tuples, where each tuple contains the column name, the window size and the function to apply.
        The function can be a predefined function (see PredefinedFunction) or a Callable (with a single parameter).
    fillna: bool = False
        Fill NaNs in the resulting DataFrame



    Examples
    --------
    ```pycon
    >>> from fold.loop import train_backtest
    >>> from fold.splitters import SlidingWindowSplitter
    >>> from fold.transformations import AddWindowFeatures
    >>> from fold.utils.tests import generate_sine_wave_data
    >>> X, y  = generate_sine_wave_data()
    >>> splitter = SlidingWindowSplitter(train_window=0.5, step=0.2)
    >>> pipeline = AddWindowFeatures(("sine", 10, "mean"))
    >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)
    >>> preds.head()
                           sine  sine~mean_10
    2021-12-31 15:40:00 -0.0000      -0.05649
    2021-12-31 15:41:00  0.0126      -0.04394
    2021-12-31 15:42:00  0.0251      -0.03139
    2021-12-31 15:43:00  0.0377      -0.01883
    2021-12-31 15:44:00  0.0502      -0.00628

    ```
    """

    def __init__(
        self,
        column_window_func: Union[ColumnWindowFunction, List[ColumnWindowFunction]],
        fillna: bool = False,
        keep_original: bool = True,
        batch_columns: Optional[int] = None,
        output_dtype: Optional[type] = None,
        name: Optional[str] = None,
        params_to_try: Optional[dict] = None,
    ) -> None:
        def replace_nan(value: Optional[int], replacement: int = 0) -> int:
            return value if value is not None else replacement

        self.column_window_func = [
            (
                wrap_in_list(column),
                window,
                function
                if isinstance(function, Callable)
                else PredefinedFunction.from_str(function),
            )
            for column, window, function in wrap_in_list(column_window_func)
        ]
        max_memory = max(
            [replace_nan(window) for _, window, _ in self.column_window_func]
        )
        self.properties = Transformation.Properties(
            requires_X=True, memory_size=max_memory, disable_memory=True
        )
        self.fillna = fillna
        self.keep_original = keep_original
        self.batch_columns = batch_columns
        self.output_dtype = output_dtype
        self.params_to_try = params_to_try
        self.name = name or "AddWindowFeatures"

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        def convert_dtype_if_needed(df: pd.DataFrame) -> pd.DataFrame:
            return df.astype(self.output_dtype) if self.output_dtype is not None else df

        def apply_function(
            columns: List[str],
            window: int,
            function: Callable,
        ):
            if isinstance(function, PredefinedFunction):
                function = getattr(pd.core.window.rolling.Rolling, function.value)
            function_name = (
                function.__name__ if function.__name__ != "<lambda>" else "transformed"
            )

            def function_to_apply(df: pd.DataFrame) -> pd.DataFrame:
                if window is None:
                    df = df.add_suffix(
                        f"{feature_name_separator}{function_name}_expanding"
                    ).expanding()
                else:
                    df = df.add_suffix(
                        f"{feature_name_separator}{function_name}_{window}"
                    ).rolling(window, min_periods=1)
                return function(df)

            return convert_dtype_if_needed(
                apply_function_batched(
                    X[get_list_column_names(columns, X)],
                    function_to_apply,
                    self.batch_columns,
                )
            )

        X_function_applied = [
            apply_function(columns, window, function)
            for columns, window, function in self.column_window_func
        ]
        to_concat = (
            [X] + X_function_applied if self.keep_original else X_function_applied
        )
        concatenated = pd.concat(to_concat, copy=False, axis="columns")

        return fill_na_inf(concatenated) if self.fillna else concatenated

    fit = fit_noop
    update = fit


ColumnFunction = Tuple[ColumnOrColumns, Callable]


class AddFeatures(Transformation, Tunable):
    """
    Applies a function to one or more columns.

    Parameters
    ----------
    column_func: Tuple[Union[str, List[str]], Callable]
        A tuple of a column or list of columns and a function to apply to them.
    fillna: bool = False
        Fill NaNs in the resulting DataFrame
    name: str
        Name of the transformation.
    params_to_try: dict
        Dictionary of parameters to try when tuning.

    Returns
    ----------
    Tuple[pd.DataFrame, Optional[Artifact]]: returns the transformed DataFrame with the original dataframe concatinated.

    Examples
    --------
    ```pycon
    >>> from fold.loop import train_backtest
    >>> from fold.splitters import SlidingWindowSplitter
    >>> from fold.transformations import AddFeatures
    >>> from fold.models.dummy import DummyClassifier
    >>> from fold.utils.tests import generate_sine_wave_data
    >>> import numpy as np
    >>> X, y  = generate_sine_wave_data()
    >>> splitter = SlidingWindowSplitter(train_window=0.5, step=0.2)
    >>> pipeline = AddFeatures([("sine", np.square)])
    >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)
    >>> preds.head()
                           sine  sine~square
    2021-12-31 15:40:00 -0.0000     0.000000
    2021-12-31 15:41:00  0.0126     0.000159
    2021-12-31 15:42:00  0.0251     0.000630
    2021-12-31 15:43:00  0.0377     0.001421
    2021-12-31 15:44:00  0.0502     0.002520

    ```


    """

    def __init__(
        self,
        column_func: Union[ColumnFunction, List[ColumnFunction]],
        past_window_size: Optional[int] = None,
        fillna: bool = False,
        keep_original: bool = True,
        batch_columns: Optional[int] = None,
        output_dtype: Optional[type] = None,
        name: Optional[str] = None,
        params_to_try: Optional[dict] = None,
    ) -> None:
        self.column_func = [
            (wrap_in_list(column), function)
            for column, function in wrap_in_list(column_func)
        ]

        self.properties = Transformation.Properties(
            requires_X=True,
            memory_size=past_window_size,
            disable_memory=True,
        )
        self.fillna = fillna
        self.keep_original = keep_original
        self.batch_columns = batch_columns
        self.output_dtype = output_dtype
        self.params_to_try = params_to_try
        self.name = name or f"ApplyFunction_{self.column_func}"

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        def convert_dtype_if_needed(df: pd.DataFrame) -> pd.DataFrame:
            return df.astype(self.output_dtype) if self.output_dtype is not None else df

        def apply_function(columns: List[str], function: Callable) -> pd.DataFrame:
            function_name = (
                function.__name__ if function.__name__ != "<lambda>" else "transformed"
            )

            return convert_dtype_if_needed(
                apply_function_batched(
                    X[get_list_column_names(columns, X)].add_suffix(
                        f"{feature_name_separator}{function_name}"
                    ),
                    function,
                    self.batch_columns,
                )
            )

        X_function_applied = [
            apply_function(columns, function) for columns, function in self.column_func
        ]
        to_concat = (
            [X] + X_function_applied if self.keep_original else X_function_applied
        )
        concatenated = pd.concat(to_concat, copy=False, axis="columns")
        return fill_na_inf(concatenated) if self.fillna else concatenated

    fit = fit_noop
    update = fit


class AddRollingCorrelation(Transformation, Tunable):
    def __init__(
        self,
        column_pairs: Union[Tuple[str], List[Tuple[str]]],
        window: int,
        fillna: bool = False,
        keep_original: bool = True,
        output_dtype: Optional[type] = None,
        name: Optional[str] = None,
        params_to_try: Optional[dict] = None,
    ) -> None:
        self.column_pairs = wrap_in_list(column_pairs)
        self.fillna = fillna
        assert all(
            len(pair) == 2 for pair in self.column_pairs
        ), "All column pairs must be of length 2"
        self.window = window
        self.properties = Transformation.Properties(
            requires_X=True, memory_size=window, disable_memory=True
        )
        self.keep_original = keep_original
        self.output_dtype = output_dtype
        self.params_to_try = params_to_try
        self.name = name or "AddRollingCorrelation"

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        def convert_dtype_if_needed(df: pd.DataFrame) -> pd.DataFrame:
            return df.astype(self.output_dtype) if self.output_dtype is not None else df

        def apply_function(
            column_pair: Tuple[str],
            window: int,
        ):
            lhs = column_pair[0]
            rhs = column_pair[1]
            return convert_dtype_if_needed(
                X[lhs]
                .rolling(window, min_periods=1)
                .corr(X[rhs])
                .rename(f"{lhs}_{rhs}{feature_name_separator}rolling_corr_{window}")
            )

        X_function_applied = [
            apply_function(column_pair, self.window)
            for column_pair in self.column_pairs
        ]
        to_concat = (
            [X] + X_function_applied if self.keep_original else X_function_applied
        )
        concatenated = pd.concat(to_concat, copy=False, axis="columns")
        return fill_na_inf(concatenated) if self.fillna else concatenated

    fit = fit_noop
    update = fit



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from importlib.util import find_spec
from typing import Callable, List, Tuple

from sklearn.base import ClassifierMixin, RegressorMixin, TransformerMixin
from sklearn.feature_selection import SelectorMixin

from ..base import Clonable, Pipeline, Transformation
from ..models.sklearn import WrapSKLearnClassifier, WrapSKLearnRegressor
from ..models.wrappers.convenience import (
    _wrap_lightgbm,
    _wrap_prophet,
    _wrap_sktime,
    _wrap_statsforecast,
    _wrap_xgboost,
)
from ..transformations.function import ApplyFunction
from ..transformations.sklearn import (
    WrapSKLearnFeatureSelector,
    WrapSKLearnTransformation,
)


def wrap_transformation_if_needed(
    transformation: Pipeline,
) -> Pipeline:
    if isinstance(transformation, List) or isinstance(transformation, Tuple):
        return [wrap_transformation_if_needed(t) for t in transformation]
    elif find_spec("xgboost") is not None and _wrap_xgboost(transformation) is not None:
        return _wrap_xgboost(transformation)  # type: ignore (we already check if it's not None)
    elif (
        find_spec("lightgbm") is not None and _wrap_lightgbm(transformation) is not None
    ):
        return _wrap_lightgbm(transformation)  # type: ignore
    elif find_spec("prophet") is not None and _wrap_prophet(transformation) is not None:
        return _wrap_prophet(transformation)  # type: ignore
    elif find_spec("sktime") is not None and _wrap_sktime(transformation) is not None:
        return _wrap_sktime(transformation)  # type: ignore
    elif (
        find_spec("statsforecast") is not None
        and _wrap_statsforecast(transformation) is not None
    ):
        return _wrap_statsforecast(transformation)  # type: ignore
    elif isinstance(transformation, RegressorMixin):
        return WrapSKLearnRegressor.from_model(transformation)
    elif isinstance(transformation, ClassifierMixin):
        return WrapSKLearnClassifier.from_model(transformation)
    elif isinstance(transformation, Callable):
        return ApplyFunction(transformation, None)
    elif isinstance(transformation, SelectorMixin):
        return WrapSKLearnFeatureSelector.from_model(transformation)
    elif isinstance(transformation, TransformerMixin):
        return WrapSKLearnTransformation.from_model(transformation)
    elif isinstance(transformation, Clonable):
        return transformation.clone(wrap_transformation_if_needed)
    elif isinstance(transformation, Transformation):
        return transformation
    else:
        raise ValueError(f"Transformation {transformation} is not supported")



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from __future__ import annotations

import logging
from typing import List, Optional, Tuple, TypeVar, Union

import pandas as pd
from tqdm import tqdm

from ..base import (
    Artifact,
    Composite,
    Optimizer,
    Pipeline,
    Sampler,
    TrainedPipeline,
    TrainedPipelines,
    Transformation,
    X,
)
from ..models.base import Model
from ..splitters import Fold
from ..utils.checks import is_prediction
from ..utils.dataframe import concat_on_columns
from ..utils.list import unpack_list_of_tuples
from ..utils.trim import trim_initial_nans_single
from .process.process_inner_loop import _process_with_inner_loop
from .process.process_minibatch import (
    _process_internal_online_model_minibatch_inference_and_update,
    _process_minibatch_transformation,
)
from .types import Backend, Stage
from .utils import (
    _cut_to_backtesting_window,
    _cut_to_train_window,
    _extract_trained_pipelines,
    _set_metadata,
    deepcopy_pipelines,
    replace_with,
)

logger = logging.getLogger("fold:loop")
DEBUG_MULTI_PROCESSING = False

T = TypeVar(
    "T",
    bound=Union[
        Transformation,
        Composite,
        Optimizer,
        Sampler,
        List[Union[Transformation, Optimizer, Sampler, Composite]],
    ],
)


def __post_checks(
    pipeline: T, X: pd.DataFrame, artifacts: Artifact
) -> Tuple[T, X, Artifact]:
    assert X.shape[0] == artifacts.shape[0]
    assert X.index.equals(artifacts.index)
    return pipeline, X, artifacts


def recursively_transform(
    X: X,
    y: Optional[pd.Series],
    artifacts: Artifact,
    transformations: T,
    stage: Stage,
    backend: Backend,
    tqdm: Optional[tqdm] = None,
) -> Tuple[T, X, Artifact]:
    """
    The main function to transform (and fit or update) a pipline of transformations.
    `stage` is used to determine whether to run the inner loop for online models.
    """
    logger.debug(f"Processing {transformations.__class__.__name__} with stage {stage}")

    if tqdm is not None and hasattr(transformations, "name"):
        tqdm.set_description(f"Processing: {transformations.name}")

    if isinstance(transformations, List) or isinstance(transformations, Tuple):
        processed_transformations = []
        for transformation in transformations:
            processed_transformation, X, artifacts = recursively_transform(
                X,
                y,
                artifacts,
                transformation,
                stage,
                backend,
                tqdm,
            )
            processed_transformations.append(processed_transformation)
        return __post_checks(processed_transformations, X, artifacts)

    elif isinstance(transformations, Composite):
        return __post_checks(
            *_process_composite(
                transformations,
                X,
                y,
                artifacts,
                stage,
                backend,
                tqdm,
            )
        )
    elif isinstance(transformations, Optimizer):
        return __post_checks(
            *_process_optimizer(
                transformations,
                X,
                y,
                artifacts,
                stage,
                backend,
                tqdm,
            )
        )

    elif isinstance(transformations, Sampler):
        return __post_checks(
            *_process_sampler(
                transformations,
                X,
                y,
                artifacts,
                stage,
                backend,
                tqdm,
            )
        )

    elif isinstance(transformations, Transformation) or isinstance(
        transformations, Model
    ):
        # If the transformation needs to be "online", and we're in the update stage, we need to run the inner loop.
        if (
            transformations.properties.mode == Transformation.Properties.Mode.online
            and stage in [Stage.update, Stage.update_online_only]
            and not transformations.properties._internal_supports_minibatch_backtesting
        ):
            return __post_checks(
                *_process_with_inner_loop(transformations, X, y, artifacts)
            )
        # If the transformation is "online" but also supports our internal "mini-batch"-style updating
        elif (
            transformations.properties.mode == Transformation.Properties.Mode.online
            and stage in [Stage.update, Stage.update_online_only]
            and transformations.properties._internal_supports_minibatch_backtesting
        ):
            return __post_checks(
                *_process_internal_online_model_minibatch_inference_and_update(
                    transformations, X, y, artifacts
                )
            )

        # or perform "mini-batch" updating OR the initial fit.
        else:
            return __post_checks(
                *_process_minibatch_transformation(
                    transformations,
                    X,
                    y,
                    artifacts,
                    stage,
                )
            )

    else:
        raise ValueError(
            f"{transformations} is not a Fold Transformation, but of type"
            f" {type(transformations)}"
        )


def _process_composite(
    composite: Composite,
    X: pd.DataFrame,
    y: Optional[pd.Series],
    artifacts: Artifact,
    stage: Stage,
    backend: Backend,
    tqdm: Optional[tqdm] = None,
) -> Tuple[Composite, X, Artifact]:
    composite.before_fit(X)
    primary_transformations = composite.get_children_primary()

    (
        primary_transformations,
        results_primary,
        y_primary,
        artifacts_primary,
    ) = unpack_list_of_tuples(
        backend.process_child_transformations(
            __process_primary_child_transform,
            enumerate(primary_transformations),
            composite,
            X,
            y,
            artifacts,
            stage,
            backend,
            None,
            tqdm,
        )
    )
    if composite.properties.artifacts_length_should_match:
        assert all(
            [
                r.shape[0] == a.shape[0]
                for r, a in zip(results_primary, artifacts_primary)
            ]
        ), ValueError("Artifacts shape doesn't match result's length.")
    composite = composite.clone(replace_with(primary_transformations))

    if composite.properties.primary_only_single_pipeline:
        assert len(results_primary) == 1, ValueError(
            "Expected single output from primary transformations, got"
            f" {len(results_primary)} instead."
        )
    if composite.properties.primary_requires_predictions:
        assert is_prediction(results_primary[0]), ValueError(
            "Expected predictions from primary transformations, but got something else."
        )

    secondary_transformations = composite.get_children_secondary()

    original_results_primary = results_primary
    results_primary = composite.postprocess_result_primary(
        results=results_primary,
        y=y_primary[0],
        original_artifact=artifacts,
        fit=stage.is_fit_or_update(),
    )
    artifacts_primary = composite.postprocess_artifacts_primary(
        primary_artifacts=artifacts_primary,
        results=original_results_primary,
        fit=stage.is_fit_or_update(),
        original_artifact=artifacts,
    )
    if composite.properties.artifacts_length_should_match:
        assert artifacts_primary.shape[0] == results_primary.shape[0], ValueError(
            f"Artifacts shape doesn't match result's length after {composite.__class__.__name__}.postprocess_artifacts_primary() was called"
        )
    if secondary_transformations is None:
        return (
            composite,
            results_primary,
            artifacts_primary,
        )

    (
        secondary_transformations,
        results_secondary,
        artifacts_secondary,
    ) = unpack_list_of_tuples(
        backend.process_child_transformations(
            __process_secondary_child_transform,
            enumerate(secondary_transformations),
            composite,
            X,
            y,
            artifacts,
            stage,
            backend,
            results_primary,
            tqdm,
        )
    )
    composite = composite.clone(replace_with(secondary_transformations))

    if composite.properties.secondary_only_single_pipeline:
        assert len(results_secondary) == 1, ValueError(
            "Expected single output from secondary transformations, got"
            f" {len(results_secondary)} instead."
        )
    if composite.properties.secondary_requires_predictions:
        assert is_prediction(results_secondary[0]), ValueError(
            "Expected predictions from secondary transformations, but got"
            " something else."
        )

    return (
        composite,
        composite.postprocess_result_secondary(
            results_primary,
            results_secondary,
            y,
            in_sample=stage == Stage.inital_fit,
        ),
        composite.postprocess_artifacts_secondary(
            artifacts_primary, artifacts_secondary, artifacts
        ),
    )


def _process_sampler(
    sampler: Sampler,
    X: pd.DataFrame,
    y: Optional[pd.Series],
    artifacts: Artifact,
    stage: Stage,
    backend: Backend,
    tqdm: Optional[tqdm] = None,
) -> Tuple[Composite, X, Artifact]:
    primary_transformations = sampler.get_children_primary()

    (
        primary_transformations,
        primary_results,
        y_primary,
        primary_artifacts,
    ) = unpack_list_of_tuples(
        backend.process_child_transformations(
            __process_primary_child_transform,
            enumerate(primary_transformations),
            sampler,
            X,
            y,
            artifacts,
            stage,
            backend,
            None,
            tqdm,
        )
    )
    sampler = sampler.clone(replace_with(primary_transformations))

    assert len(primary_results) == 1, ValueError(
        "Expected single output from primary transformations, got"
        f" {len(primary_results)} instead."
    )

    if stage is Stage.inital_fit:
        secondary_transformations = sampler.get_children_primary()
        (
            secondary_transformations,
            primary_results,
            primary_y,
            primary_artifacts,
        ) = unpack_list_of_tuples(
            backend.process_child_transformations(
                __process_primary_child_transform,
                enumerate(secondary_transformations),
                sampler,
                X,
                y,
                artifacts,
                Stage.infer,
                backend,
                None,
                tqdm,
            )
        )

    primary_results = primary_results[0]
    primary_artifacts = primary_artifacts[0]
    assert primary_artifacts.shape[0] == primary_results.shape[0], ValueError(
        f"Artifacts shape doesn't match result's length after {sampler.__class__.__name__}.postprocess_artifacts_primary() was called"
    )
    return sampler, primary_results, primary_artifacts


def _process_optimizer(
    optimizer: Optimizer,
    X: pd.DataFrame,
    y: Optional[pd.Series],
    artifacts: Artifact,
    stage: Stage,
    backend: Backend,
    tqdm: Optional[tqdm] = None,
) -> Tuple[Pipeline, X, Artifact]:
    if tqdm is not None:
        tqdm.set_description(f"Processing: {optimizer.name}")
    optimized_pipeline = optimizer.get_optimized_pipeline()
    artifact = None
    if optimized_pipeline is None:
        while True:
            candidates = optimizer.get_candidates()
            if len(candidates) == 0:
                break

            _, results, candidate_artifacts = unpack_list_of_tuples(
                backend.process_child_transformations(
                    __process_candidates,
                    enumerate(candidates),
                    optimizer,
                    X,
                    y,
                    artifacts,
                    stage,
                    backend,
                    None,
                    None,
                )
            )
            results = [trim_initial_nans_single(result) for result in results]
            artifact = optimizer.process_candidate_results(
                results,
                y=y.loc[results[0].index],
                artifacts=candidate_artifacts,
            )

    optimized_pipeline = optimizer.get_optimized_pipeline()
    processed_optimized_pipeline, X, artifact = recursively_transform(
        X,
        y,
        concat_on_columns([artifact, artifacts], copy=False),
        optimized_pipeline,
        stage,
        backend,
    )
    return optimizer, X, artifact


def __process_candidates(
    optimizer: Optimizer,
    index: int,
    child_transform: Pipeline,
    X: pd.DataFrame,
    y: Optional[pd.Series],
    artifacts: Artifact,
    stage: Stage,
    backend: Backend,
    results_primary: Optional[List[pd.DataFrame]],
    tqdm: Optional[tqdm] = None,
) -> Tuple[Pipeline, X, Artifact]:
    splits = optimizer.splitter.splits(y.index)

    (
        processed_idx,
        processed_pipelines,
        processed_predictions,
        processed_artifacts,
    ) = _sequential_train_on_window(
        child_transform,
        X,
        y,
        splits,
        artifacts,
        backend=backend,
        project_name=f"HPO-{optimizer.name}",
    )
    trained_pipelines = _extract_trained_pipelines(processed_idx, processed_pipelines)

    result, artifact = _backtest_on_window(
        trained_pipelines,
        splits[0],
        X,
        y,
        artifacts,
        backend,
        mutate=False,
    )
    assert result.index.equals(artifact.index)
    return (
        trained_pipelines,
        trim_initial_nans_single(result),
        artifact,
    )


def __process_primary_child_transform(
    composite: Union[Composite, Sampler],
    index: int,
    child_transform: Pipeline,
    X: pd.DataFrame,
    y: Optional[pd.Series],
    artifacts: Artifact,
    stage: Stage,
    backend: Backend,
    results_primary: Optional[List[pd.DataFrame]],
    tqdm: Optional[tqdm] = None,
) -> Tuple[Pipeline, X, Optional[pd.Series], Artifact]:
    X, y, artifacts = composite.preprocess_primary(
        X=X, index=index, y=y, artifact=artifacts, fit=stage.is_fit_or_update()
    )
    transformations, X, artifacts = recursively_transform(
        X,
        y,
        artifacts,
        child_transform,
        stage,
        backend,
        tqdm,
    )
    return transformations, X, y, artifacts


def __process_secondary_child_transform(
    composite: Composite,
    index: int,
    child_transform: Pipeline,
    X: pd.DataFrame,
    y: Optional[pd.Series],
    artifacts: Artifact,
    stage: Stage,
    backend: Backend,
    results_primary: Optional[List[pd.DataFrame]],
    tqdm: Optional[tqdm] = None,
) -> Tuple[Pipeline, X, Artifact]:
    X, y, artifacts = composite.preprocess_secondary(
        X=X,
        y=y,
        artifact=artifacts,
        results_primary=results_primary,
        index=index,
        fit=stage.is_fit_or_update(),
    )
    return recursively_transform(
        X,
        y,
        artifacts,
        child_transform,
        stage,
        backend,
        tqdm,
    )


def _backtest_on_window(
    trained_pipelines: TrainedPipelines,
    split: Fold,
    X: pd.DataFrame,
    y: pd.Series,
    artifact: Artifact,
    backend: Backend,
    mutate: bool,
) -> Tuple[X, Artifact]:
    pd.options.mode.copy_on_write = True
    current_pipeline = [
        pipeline_over_time.loc[split.model_index]
        for pipeline_over_time in trained_pipelines
    ]
    if not mutate:
        current_pipeline = deepcopy_pipelines(current_pipeline)

    X_test = _cut_to_backtesting_window(X, split, current_pipeline)
    y_test = _cut_to_backtesting_window(y, split, current_pipeline)
    artifact_test = _cut_to_backtesting_window(artifact, split, current_pipeline)

    original_idx = X_test.index
    X_test, y_test, artifact_test = _get_X_y_based_on_events(
        X_test, y_test, artifact_test
    )
    artifacts_to_check = Artifact.get_events(artifact_test)
    if artifacts_to_check is None:
        artifacts_to_check = artifact_test
    assert artifacts_to_check.dropna().shape[0] == y_test.shape[0]

    results, artifacts = recursively_transform(
        X=X_test,
        y=y_test,
        artifacts=artifact_test,
        transformations=current_pipeline,
        stage=Stage.update_online_only,
        backend=backend,
    )[1:]
    if len(results.index) != len(original_idx):
        results = results.reindex(original_idx)
        artifacts = artifacts.reindex(original_idx)
    return (
        results.loc[X.index[split.test_window_start] :],
        artifacts.loc[X.index[split.test_window_start] :],
    )


def _train_on_window(
    X: pd.DataFrame,
    y: pd.Series,
    artifact: Artifact,
    pipeline: Pipeline,
    split: Fold,
    never_update: bool,
    backend: Backend,
    project_name: str,
    show_progress: bool = False,
) -> Tuple[int, TrainedPipeline, X, Artifact]:
    pd.options.mode.copy_on_write = True
    stage = Stage.inital_fit if (split.order == 0 or never_update) else Stage.update

    X_train: pd.DataFrame = _cut_to_train_window(X, split, stage)
    y_train = _cut_to_train_window(y, split, stage)
    if y_train.nunique() == 1:
        raise ValueError(
            "Only a single unique value in `y_train`. This means models can't learn anything."
        )

    artifact_train = _cut_to_train_window(artifact, split, stage)

    original_idx = X_train.index
    X_train, y_train, artifact_train = _get_X_y_based_on_events(
        X_train, y_train, artifact_train
    )

    pipeline = deepcopy_pipelines(pipeline)
    pipeline = _set_metadata(
        pipeline,
        Composite.Metadata(
            project_name=project_name,
            fold_index=split.order,
            target=y.name,
            inference=False,
        ),
    )
    trained_pipeline, X_train, artifact_train = recursively_transform(
        X=X_train,
        y=y_train,
        artifacts=artifact_train,
        transformations=pipeline,
        stage=stage,
        backend=backend,
        tqdm=tqdm() if show_progress else None,
    )
    if len(X_train.index) != len(original_idx):
        X_train = X_train.reindex(original_idx)
        artifact_train = artifact_train.reindex(original_idx)

    return split.model_index, trained_pipeline, X_train, artifact_train


def _sequential_train_on_window(
    pipeline: Pipeline,
    X: Optional[pd.DataFrame],
    y: pd.Series,
    splits: List[Fold],
    artifact: Artifact,
    backend: Backend,
    project_name: str,
) -> Tuple[List[int], List[Pipeline], List[X], List[Artifact]]:
    processed_idx = []
    processed_pipelines: List[Pipeline] = []
    processed_pipeline = pipeline
    processed_predictions = []
    processed_artifacts = []
    for split in splits:
        (
            processed_id,
            processed_pipeline,
            processed_prediction,
            processed_artifact,
        ) = _train_on_window(
            X,
            y,
            artifact,
            processed_pipeline,
            split,
            never_update=False,
            backend=backend,
            project_name=project_name,
        )
        processed_idx.append(processed_id)
        processed_pipelines.append(processed_pipeline)
        processed_predictions.append(processed_prediction)
        processed_artifacts.append(processed_artifact)

    return (
        processed_idx,
        processed_pipelines,
        processed_predictions,
        processed_artifacts,
    )


def _get_X_y_based_on_events(
    X: pd.DataFrame,
    y: pd.Series,
    artifact: Artifact,
) -> Tuple[pd.DataFrame, pd.Series, Artifact]:
    events = Artifact.get_events(artifact)
    if events is None:
        return X, y, artifact
    events = events.dropna()
    assert len(events) > 0, "No events found in fold"
    return X.loc[events.index], events.event_label, artifact.loc[events.index]



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.

from __future__ import annotations

from typing import Optional, Union

import pandas as pd

from ..base import Artifact, InvertibleTransformation, Transformation, Tunable
from ..utils.dataframe import fill_na_inf, take_log, to_series


class Difference(InvertibleTransformation, Tunable):
    """
    Performs differencing.
    Sesonal differencing can be achieved by setting `lag` to the seasonality of the data.
    To achieve second-order differencing, simply chain multiple `Difference` transformations.

    Parameters
    ----------
    lag : int, optional
        the seasonality of the data, by default 1

    Examples
    --------
    ```pycon
    >>> from fold.loop import train_backtest
    >>> from fold.splitters import SlidingWindowSplitter
    >>> from fold.transformations import Difference
    >>> from fold.utils.tests import generate_sine_wave_data
    >>> X, y  = generate_sine_wave_data(freq="min")
    >>> splitter = SlidingWindowSplitter(train_window=0.5, step=0.2)
    >>> pipeline = Difference()
    >>> X["sine"].head()
    2021-12-31 07:20:00    0.0000
    2021-12-31 07:21:00    0.0126
    2021-12-31 07:22:00    0.0251
    2021-12-31 07:23:00    0.0377
    2021-12-31 07:24:00    0.0502
    Freq: T, Name: sine, dtype: float64
    >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)
    >>> preds["sine"].head()
    2021-12-31 15:40:00    0.0126
    2021-12-31 15:41:00    0.0126
    2021-12-31 15:42:00    0.0125
    2021-12-31 15:43:00    0.0126
    2021-12-31 15:44:00    0.0125
    Freq: T, Name: sine, dtype: float64

    ```

    References
    ----------

    [Stationarity and differencing](https://otexts.com/fpp2/stationarity.html)
    """

    first_values_X: Optional[Union[pd.DataFrame, pd.Series]] = None
    last_values_X: Optional[Union[pd.DataFrame, pd.Series]] = None

    def __init__(
        self,
        lag: int = 1,
        name: Optional[str] = None,
        params_to_try: Optional[dict] = None,
    ) -> None:
        self.lag = lag
        self.params_to_try = params_to_try
        self.name = name or "Difference"
        self.properties = InvertibleTransformation.Properties(requires_X=False)

    def fit(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        sample_weights: Optional[pd.Series] = None,
    ) -> Optional[Artifact]:
        self.last_values_X = X.iloc[-self.lag : None]
        self.first_values_X = X.iloc[: self.lag]

    def update(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        sample_weights: Optional[pd.Series] = None,
    ) -> Optional[Artifact]:
        self.last_values_X = X.iloc[-self.lag : None]

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        if in_sample:
            return X.diff(self.lag).fillna(0.0)
        else:
            return (
                pd.concat([self.last_values_X.copy(), X], copy=False, axis="index")
                .diff(self.lag)
                .iloc[self.lag :]
                .fillna(0.0)
            )

    def inverse_transform(self, X: pd.Series, in_sample: bool) -> pd.Series:
        if in_sample:
            X = X.copy()
            X.iloc[: self.lag] = to_series(self.first_values_X)
            for i in range(self.lag):
                X.iloc[i :: self.lag] = X.iloc[i :: self.lag].cumsum()
            return X
        else:
            X = pd.concat(
                [to_series(self.last_values_X).copy(), X], copy=False, axis="index"
            )
            for i in range(self.lag):
                X.iloc[i :: self.lag] = X.iloc[i :: self.lag].cumsum()
            return X.iloc[self.lag :]


class TakeReturns(Transformation, Tunable):
    """
    Takes the returns (percentage change between the current and a prior element).

    Parameters
    ----------
    log_returns : bool, optional, default False.
        If True, computes the log returns instead of the simple returns, default False.

    Examples
    --------
    ```pycon
    >>> from fold.loop import train_backtest
    >>> from fold.splitters import SlidingWindowSplitter
    >>> from fold.transformations import TakeReturns
    >>> from fold.utils.tests import generate_sine_wave_data
    >>> X, y  = generate_sine_wave_data(freq="min")
    >>> splitter = SlidingWindowSplitter(train_window=0.5, step=0.2)
    >>> pipeline = TakeReturns()
    >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)
    >>> X["sine"].loc[preds.index].head()
    2021-12-31 15:40:00   -0.0000
    2021-12-31 15:41:00    0.0126
    2021-12-31 15:42:00    0.0251
    2021-12-31 15:43:00    0.0377
    2021-12-31 15:44:00    0.0502
    Freq: T, Name: sine, dtype: float64
    >>> preds["sine"].head()
    2021-12-31 15:40:00   -1.000000
    2021-12-31 15:41:00        -inf
    2021-12-31 15:42:00    0.992063
    2021-12-31 15:43:00    0.501992
    2021-12-31 15:44:00    0.331565
    Freq: T, Name: sine, dtype: float64

    ```

    """

    last_values_X: Optional[Union[pd.DataFrame, pd.Series]] = None

    def __init__(
        self,
        log_returns: bool = False,
        fillna: bool = False,
        name: Optional[str] = None,
        params_to_try: Optional[dict] = None,
    ) -> None:
        self.log_returns = log_returns
        self.fillna = fillna
        self.params_to_try = params_to_try
        description = "log" if log_returns else "arithmetic"
        self.name = name or f"TakeReturns-{description}"
        self.properties = InvertibleTransformation.Properties(requires_X=False)

    def fit(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        sample_weights: Optional[pd.Series] = None,
    ) -> Optional[Artifact]:
        self.last_values_X = X.iloc[-1:None]

    def update(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        sample_weights: Optional[pd.Series] = None,
    ) -> Optional[Artifact]:
        self.last_values_X = X.iloc[-1:None]

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        def operation(df):
            if self.log_returns:
                return take_log(df / df.shift(1))
            else:
                return df.pct_change()

        fill_na = fill_na_inf if self.fillna else lambda x: x

        if in_sample:
            return fill_na(operation(X))
        else:
            return fill_na(
                operation(
                    pd.concat([self.last_values_X.copy(), X], copy=False, axis="index")
                )
            ).iloc[1:]



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from random import choices
from typing import Callable, List, Optional, Tuple

import numpy as np
import pandas as pd

from fold.base import Transformation, Tunable
from fold.loop import train_backtest
from fold.splitters import SingleWindowSplitter


def generate_sine_wave_data(
    cycles: int = 2, length: int = 1000, freq: str = "min"
) -> Tuple[pd.DataFrame, pd.Series]:
    end_value = np.pi * 2 * cycles
    my_wave = np.sin(np.linspace(0, end_value, length + 1))
    series = pd.Series(
        my_wave,
        name="sine",
        index=pd.date_range(end="2022", periods=len(my_wave), freq=freq),
    ).round(4)
    X = series.to_frame()
    y = series.shift(-1)[:-1]
    X = X[:-1]
    return X, y


def generate_all_zeros(length: int = 1000) -> Tuple[pd.DataFrame, pd.Series]:
    length += 1
    series = pd.Series(
        [0] * length,
        index=pd.date_range(end="2022", periods=length, freq="m"),
    )
    X = series.to_frame()
    y = series.shift(-1)[:-1]
    X = X[:-1]
    return X, y


def generate_zeros_and_ones(
    length: int = 1000, labels=[1, 0]
) -> Tuple[pd.DataFrame, pd.Series]:
    return generate_zeros_and_ones_skewed(length, labels, weights=[0.5, 0.5])


def generate_zeros_and_ones_skewed(
    length: int = 1000, labels=[1, 0], weights: List[float] = [0.2, 0.8]
) -> Tuple[pd.DataFrame, pd.Series]:
    length += 1
    series = pd.Series(
        choices(population=labels, weights=weights, k=length),
        index=pd.date_range(end="2022", periods=length, freq="s"),
    )
    X = series.to_frame()
    y = series.shift(-1)[:-1]
    X = X[:-1]
    return X, y


def generate_monotonous_data(
    length: int = 1000, freq: str = "m"
) -> Tuple[pd.DataFrame, pd.Series]:
    values = np.linspace(0, 1, num=length + 1)
    series = pd.Series(
        values,
        name="linear",
        index=pd.date_range(end="2022", periods=len(values), freq=freq),
    )
    X = series.to_frame()
    y = series.shift(-1)[:-1]
    X = X[:-1]
    return X, y


def check_if_transformation_mutates(
    X: pd.DataFrame, y: pd.Series, transformation: Transformation
) -> None:
    X_train = X[: len(X) // 2]
    y_train = y[: len(y) // 2]
    X_train_before = X_train.copy()
    y_train_before = y_train.copy()
    transformation.fit(X_train, y_train)
    assert X_train_before.equals(X_train)
    assert y_train_before.equals(y_train)

    X_test = X[len(X) // 2 :]
    y_test = y[len(y) // 2 :]
    X_test_before = X_test.copy()
    y_test_before = y_test.copy()
    transformation.transform(X_test, in_sample=True)
    transformation.update(X, y)
    assert X_test_before.equals(X_test)
    assert y_test_before.equals(y_test)

    X_val = X[-1:]
    y_val = y[-1:]
    X_val_before = X_val.copy()
    y_val_before = y_val.copy()
    transformation.transform(X_val, in_sample=False)
    assert X_val_before.equals(X_val)
    assert y_val_before.equals(y_val)


def tuneability_test(
    instance: Transformation,
    different_params: dict,
    init_function: Optional[Callable] = None,
    classification: bool = False,
    tolerance: Optional[float] = None,
    modify_X_func: Optional[Callable] = None,
):
    """
    Used to test the general structure and implementation of get_params() and clone_with_params() methods.
    the kwargs are used to set the parameters of the transformation, which should be different to the init parameters of the `instance`.
    """
    assert isinstance(instance, Tunable)
    assert instance.id is not None

    params = instance.get_params()
    different_instance = (
        instance.__class__(**different_params)
        if init_function is None
        else init_function(**different_params)
    )

    X, y = (
        generate_zeros_and_ones(length=500)
        if classification
        else generate_sine_wave_data(length=500)
    )
    if modify_X_func is not None:
        X = modify_X_func(X)
    splitter = SingleWindowSplitter(train_window=0.5)
    preds_orig, _ = train_backtest(instance, X, y, splitter)
    preds_different, _ = train_backtest(different_instance, X, y, splitter)
    assert not preds_orig.equals(
        preds_different
    ), "The output of the two instances with different parameters should be different"

    reconstructed_instance = different_instance.clone_with_params(params)

    preds_reconstructed, _ = train_backtest(reconstructed_instance, X, y, splitter)
    if tolerance is None:
        assert preds_orig.equals(preds_reconstructed), (
            "The output of the two instances with different parameters should be the"
            " same"
        )
    else:
        assert np.allclose(preds_orig, preds_reconstructed, atol=tolerance), (
            "The output of the two instances with the same parameters should be the"
            " close"
        )



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


import logging
from typing import Optional, Tuple, Union

import pandas as pd

from ..base import Artifact, EventDataFrame, OutOfSamplePredictions, TrainedPipelineCard
from ..events import _create_events
from ..splitters import Fold, Splitter
from ..utils.dataframe import concat_on_index
from ..utils.list import unpack_list_of_tuples
from ..utils.trim import trim_initial_nans_single
from .backend import get_backend
from .checks import check_types
from .common import _backtest_on_window
from .types import Backend, BackendType

logger = logging.getLogger("fold:loop")


def backtest(
    trained_pipelinecard: TrainedPipelineCard,
    X: Optional[pd.DataFrame],
    y: pd.Series,
    splitter: Splitter,
    backend: Union[BackendType, Backend, str] = BackendType.no,
    sample_weights: Optional[pd.Series] = None,
    events: Optional[EventDataFrame] = None,
    silent: bool = False,
    mutate: bool = False,
    return_artifacts: bool = False,
) -> Union[OutOfSamplePredictions, Tuple[OutOfSamplePredictions, Artifact]]:
    """
    Run backtest on TrainedPipelineCard and given data.

    Parameters
    ----------

    trained_pipelines: TrainedPipelineCard
        The fitted pipelines, for all folds.
    X: pd.DataFrame, optional
        Exogenous Data.
    y: pd.Series
        Endogenous Data (Target).
    splitter: Splitter
        Defines how the folds should be constructed.
    backend: str, BackendType = BackendType.no
        The library/service to use for parallelization / distributed computing, by default `no`.
    sample_weights: pd.Series, optional = None
        Weights assigned to each sample/timestamp, that are passed into models that support it, by default None.
    events: EventDataFrame, optional = None
        Events that should be passed into the pipeline, by default None.
    silent: bool = False
        Wether the pipeline should print to the console, by default False.
    mutate: bool = False
        Whether `trained_pipelines` should be mutated, by default False. This is discouraged.
    return_artifacts: bool = False
        Whether to return the artifacts of the backtesting process, by default False.

    Returns
    -------
    OutOfSamplePredictions
        Predictions for all folds, concatenated.
    """
    backend = get_backend(backend)
    X, y = check_types(X, y)
    if events is None:
        events = _create_events(y, trained_pipelinecard)
    if events is not None and events.shape[0] != X.shape[0]:
        logger.warning("The number of events does not match the number of samples.")
        events = events.reindex(X.index)
    artifact = Artifact.from_events(X.index, events)

    if trained_pipelinecard.preprocessing is not None:
        preprocessed_X, preprocessed_artifacts = _backtest_on_window(
            trained_pipelines=trained_pipelinecard.preprocessing,
            split=Fold(0, 0, 0, len(X), 0, 0, 0, len(X)),
            X=X,
            y=y,
            artifact=artifact,
            backend=backend,
            mutate=mutate,
        )

        assert preprocessed_X.shape[0] == X.shape[0]
        assert preprocessed_artifacts.shape[0] == artifact.shape[0]
        X = preprocessed_X
        artifact = preprocessed_artifacts

    results, artifacts = unpack_list_of_tuples(
        backend.backtest_pipeline(
            _backtest_on_window,
            trained_pipelinecard.pipeline,
            splitter.splits(index=X.index),
            X,
            y,
            artifact,
            backend,
            mutate=mutate,
            silent=silent,
        )
    )
    results = trim_initial_nans_single(concat_on_index(results, copy=False))
    if return_artifacts:
        return results, concat_on_index(artifacts, copy=False)
    else:
        return results



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from typing import Union

from ..types import Backend, BackendType


def get_backend(backend_type: Union[str, BackendType, Backend]) -> Backend:
    if isinstance(backend_type, str):
        backend_type = BackendType.from_str(backend_type)

    if isinstance(backend_type, Backend):
        return backend_type

    if backend_type == BackendType.ray:
        from .ray import RayBackend

        return RayBackend()
    elif backend_type == BackendType.no:
        from .sequential import NoBackend

        return NoBackend()
    elif backend_type == BackendType.pathos:
        from .pathos import PathosBackend

        return PathosBackend()
    elif backend_type == BackendType.thread:
        from .thread import ThreadBackend

        return ThreadBackend()
    elif backend_type == BackendType.joblib:
        from .joblib import JoblibBackend

        return JoblibBackend()

    else:
        raise ValueError(f"Backend type {backend_type} not supported.")



from __future__ import annotations

from typing import Any, Optional, Union

import pandas as pd

from fold.base import fit_noop
from fold.models.base import Model
from fold.utils.checks import is_X_available


class WrapArch(Model):
    def __init__(
        self,
        init_args: dict,
        use_exogenous: Optional[bool] = None,
        online_mode: bool = False,
        instance: Optional[Any] = None,
        name: Optional[str] = None,
    ) -> None:
        self.init_args = init_args
        init_args = {} if init_args is None else init_args
        self.use_exogenous = use_exogenous
        self.properties = Model.Properties(
            requires_X=False,
            model_type=Model.Properties.ModelType.regressor,
            mode=(
                Model.Properties.Mode.online
                if online_mode
                else Model.Properties.Mode.minibatch
            ),
        )
        self.name = name or "Arch"
        self.instance = instance

    def fit(
        self, X: pd.DataFrame, y: pd.Series, sample_weights: Optional[pd.Series] = None
    ) -> None:
        from arch import arch_model

        use_exogenous = (
            is_X_available(X) if self.use_exogenous is None else self.use_exogenous
        )
        if use_exogenous:
            self.model = arch_model(y, x=X, **self.init_args)
        else:
            self.model = arch_model(y, **self.init_args)
        self.model = self.model.fit(disp="off")

    def predict(self, X: pd.DataFrame) -> Union[pd.Series, pd.DataFrame]:
        use_exogenous = (
            is_X_available(X) if self.use_exogenous is None else self.use_exogenous
        )
        if use_exogenous:
            res = self.model.forecast(horizon=len(X), reindex=False, x=X)
        else:
            res = self.model.forecast(horizon=len(X), reindex=False)
        return pd.Series(res.variance.values[0], index=X.index)

    def predict_in_sample(self, X: pd.DataFrame) -> Union[pd.Series, pd.DataFrame]:
        res = self.model.forecast(horizon=len(X), start=0, reindex=True)
        return res.variance[res.variance.columns[0]]

    update = fit_noop



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from ..base import *
from .columns import *
from .concat import *
from .ensemble import *
from .metalabeling import *
from .optimize import *
from .residual import *
from .sample import *
from .select import *
from .target import *



import numpy as np

from fold.loop import train_backtest
from fold.splitters import Splitter
from fold.utils.tests import generate_monotonous_data


def run_pipeline_and_check_if_results_close_univariate(
    model, splitter: Splitter, data_length: int = 70
):
    X, y = generate_monotonous_data(length=data_length)

    pred, _ = train_backtest(model, None, y, splitter)
    assert np.isclose(y.squeeze()[pred.index], pred.squeeze().values, atol=0.1).all()


def run_pipeline_and_check_if_results_close_exogenous(
    model, splitter: Splitter, data_length: int = 70
):
    X, y = generate_monotonous_data(length=data_length)

    pred, _ = train_backtest(model, None, y, splitter)
    assert np.isclose(y.squeeze()[pred.index], pred.squeeze().values, atol=0.1).all()
